#! /usr/bin/env python
#
#    Edgebreaker algorithm relevant pieces, imported into FVList model.
#    Copyright (C) 2002  Michael Urman
#
#    Based directly off of pseudocode from Rossignac et al.  If only it worked.
#    It's too abstract for me to debug easily, as well :(
#

import sys

class _eb_notation:
    def __init__(self, fv, face=None, index=0, flip=1):
        """_eb_notation(self, fv) -> _eb_notation

        Helps follow the edgebreaker notation from FVList models"""

        self.fv = fv
        self.face = face
        self.index = index%3
        self.flip = flip

    def __getattr__(self, attr):
        if attr == 'p':
            return _eb_notation(self.fv, self.face, (3+self.index+2*self.flip)%3)
        elif attr == 'n':
            return _eb_notation(self.fv, self.face, (3+self.index+self.flip)%3)
        elif attr == 'v':
            return self.fv.faces[self.face][self.index]
        elif attr == 't':
            return self.face
        elif attr == 'l':
            return self.n.o
        elif attr == 'r':
            return self.p.o
        elif attr == 'o':
            nv = self.n.v
            pv = self.p.v
            sf = self.fv.faces
            overtex = self.fv.ovtable[self.face][self.index]
            ofaces = [ face for face in self.fv.vtof[overtex]
                if nv in sf[face] and pv in sf[face]]
            oface = ofaces[0]
            ofv = sf[oface]
            ni = ofv.index(nv)
            pi = ofv.index(pv)
            if ni - pi in (1, -2):
                flip = self.flip
            else:
                flip = 0-self.flip
            index = [i for i in (0,1,2) if i not in (ni, pi)][0]
            return _eb_notation(self.fv, oface, index, flip)
        else:
            raise AttributeError("EB has no attribute '%s'"%attr)

def edgebreak(self, startat, meshtype='manifold'):
    """edgebreak(self, startat) -> vertex order, CLERS string
    
    Perform edgebreaker connectivity compression algorithm.  Return an
    ordered vertex list (by number) and a CLERS string (characters, not
    binary)."""

    self.build_ovtable()
    self.vtof = self.build_vtof()
    c = _eb_notation(self, startat)

    self.eb_clers = []
    self.eb_handles = []
    self.eb_M = [0 for v in self.vertices]
    self.eb_U = [0 for f in self.faces]
    self.eb_T = -1
    self.eb_vorder = []
    _eb_encode_delta(self, c.n)
    if meshtype == 'manifold':
        self.eb_M[c.n.v] = 1

    _eb_encode_delta(self, c)
    self.eb_M[c.v] = 1

    _eb_encode_delta(self, c.p)
    self.eb_M[c.p.v] = 1

    self.eb_U[c.t] = 1
    a = c.o

    count = 1
    while a.v != c.p.o.p.v:
        self.eb_U[a.t] = 1
        self.eb_T += 1
        count += 1
        _eb_encode_delta(self, a)
        self.eb_M[a.v] = 1
        a = a.n.o
    self.eb_U[a.t] = 1
    self.eb_T += 1
    count += 1
    self.eb_clers.append(meshtype)
    self.eb_clers.append(count)
    _eb_compress(self, a.p.o)

    del self.eb_M
    del self.eb_U
    del self.eb_T
    del self.vtof

def _eb_compress(self, c):
    while 1:
        self.eb_U[c.t] = 1
        self.eb_T += 1

        # handles
        if self.eb_U[c.n.o.t] > 1:
            self.eb_handles.append(self.eb_U[c.n.o.t])
            self.eb_handles.append(self.eb_T*3+1)
        if self.eb_U[c.p.o.t] > 1:
            self.eb_handles.append(self.eb_U[c.p.o.t])
            self.eb_handles.append(self.eb_T*3+2)

        if self.eb_M[c.v] != 1:
            self.eb_clers.append('C')
            _eb_encode_delta(self, c)
            self.eb_M[c.v] = 1
            c = c.r
        elif self.eb_U[c.r.t] > 0:
            if self.eb_U[c.l.t] > 0:
                self.eb_clers.append('E')
                return
            else:
                self.eb_clers.append('R')
                c = c.l
        elif self.eb_U[c.l.t] > 0:
            self.eb_clers.append('L')
            c = c.r
        else:
            self.eb_U[c.t] = self.eb_T*3+2
            self.eb_clers.append('S')
            _eb_compress(self, c.r)
            c = c.l
            if self.eb_U[c.t] > 0:
                return

def _eb_encode_delta(self, c):
    """_eb_encode_delta(self, c) -> None"""
    self.eb_vorder.append(c.v)

def save_edgebreaker(self, filebase, binary=0):
    """save_edgebreaker(self, filename, binary=0) -> None

    Saves the edgebreaker encoding to the file specified.  If the filename
    ends in .gz, the output will be gzip compressed.  If binary is 1, it
    will be encoded in binary.  This is not exclusive from gzip."""

    # first open the file for saving; this will throw any errors before we
    # do any real work.
    filename = filebase % 'clers'
    if filename.endswith(".gz"):
        try:
            import gzip
            clers = gzip.open(filename, 'wb')
        except ImportError:
            raise SystemExit("Saving gzipped file '%s' requires module gzip" % filename)
    else:
        clers = open(filename, 'wb')

    filename = filebase % 'verts'
    if filename.endswith(".gz"):
        try:
            import gzip
            verts = gzip.open(filename, 'wb')
        except ImportError:
            raise SystemExit("Saving gzipped file '%s' requires module gzip" % filename)
    else:
        verts = open(filename, 'wb')

    self.edgebreak(0)

    print ' writing clers ...',
    sys.stdout.flush()

    types = 'manifold tpatch interior_tpatch holes'.split()
    if binary:
        self.binary = []
        if self.eb_clers[0].lower() in types:
            index = types.index(self.eb_clers[0].lower())
            self.binary.append(index & 1)
            self.binary.append(index & 2)
        else:
            raise NotImplementedError("Invalid CLERS type '%s'" % self.eb_clers[0])

        for i in range(0,31):
            self.binary.append(self.eb_clers[1] & 2**i)

        optable = {'C':[0], 'L':[1,1,0], 'E':[1,1,1],
                    'R':[1,0,1], 'S':[1,0,0]}
        for op in self.eb_clers[2:]:
            self.binary.extend(optable[op])

        bits = 0
        byte = 0
        for bit in self.binary:
            byte += 2**bits * bit
            bits += 1
            if bits == 8:
                clers.write(chr(byte))
                bits = 0
                byte = 0
        if bits>0:
            clers.write(chr(byte))

    else:
        if self.eb_clers[0].lower() in types:
            clers.write('%s\n' % self.eb_clers[0].upper())
        else:
            raise NotImplementedError("Invalid CLERS type '%s'" % self.eb_clers[0])

        clers.write('%d\n' % self.eb_clers[1])

        for op in self.eb_clers[2:]:
            clers.write(op)

    print ' writing verts ...',
    sys.stdout.flush()
    for vert in self.eb_vorder:
        verts.write('%f %f %f\n' % tuple(self.vertices[vert]))

    clers.close()
    verts.close()
    del self.eb_clers
    del self.eb_handles

def load_edgebreaker(self, filebase, binary=0):
    """load_edgebreaker(self, filebase) -> None

    Loads edgebreaker model from file, replacing current model"""

    filename = filebase % 'clers'
    if filename.endswith('.gz'):
        try:
            import gzip
            self.f_clers = gzip.open(filename)
        except ImportError:
            raise SystemExit("Loading gzipped file '%s' requires module gzip" % filename)
    else:
        self.f_clers = open(filename)

    filename = filebase % 'verts'
    if filename.endswith('.gz'):
        try:
            import gzip
            self.f_verts = gzip.open(filename)
        except ImportError:
            raise SystemExit("Loading gzipped file '%s' requires module gzip" % filename)
    else:
        self.f_verts = open(filename)

    self.vertices = []
    self.faces = []
    self.ovtable = []

    V = self.eb_V = [0,2,1]
    O = self.eb_O = [-1,-1,-3]
    H = self.eb_H = []
    self.eb_T = 0
    self.eb_N = 2
    self.eb_A = 0
    if binary == 1:
        raise NotImplementedError("Can't load binary eb files")

    self.eb_meshtype = self.f_clers.readline()
    self.eb_I = int(self.f_clers.readline())
    self.eb_clers = ['C'] * (self.eb_I-2)
    self.eb_clers.append('R')
    self.eb_clers.extend(list(self.f_clers.read()))
    self.eb_clers_offset = 0
    _eb_decompress_connectivity(self, 2)

    # XXX: hmmm. 
    self.eb_M = [0 for i in range(len(self.eb_V)/3)]
    self.eb_U = [0 for i in range(self.eb_T)]
    self.eb_G = [0 for i in range(3)]

    self.eb_G[0] = _eb_decode_delta(self, 0)
    if self.eb_meshtype.lower() == 'manifold':
        self.eb_M[0] = 1

    self.eb_G[1] = _eb_decode_delta(self, 2)
    self.eb_M[1] = 1

    self.eb_G[2] = _eb_decode_delta(self, 1)
    self.eb_M[1] = 1
    self.eb_N = 2
    self.eb_U[0] = 1
    _eb_decompress_vertices(self, self.eb_O[2])

    # clean up from different format
    self.faces = []
    while len(self.eb_V):
        self.faces.append(self.eb_V[0:3])
        self.eb_V = self.eb_V[3:]
    self.eb_G.append(_eb_decode_delta(self, 1)) # XXX workaround ?!
    self.vertices = self.eb_G

    print 'faces', self.faces
    print 'vertices', self.vertices


def _eb_decompress_connectivity(self, c):
    def P(c): return c/3+(2+c)%3
    def N(c): return c/3+(1+c)%3
    O, V = self.eb_O, self.eb_V

    while 1:
        self.eb_T += 1
        V.extend([0,0,0])
        O.extend([-3,-3,-3])
        O[c] = 3*self.eb_T; O[3*self.eb_T] = c
        V[3*self.eb_T+1] = V[P(c)]
        V[3*self.eb_T+2] = V[N(c)]

        c = N(O[c])

        clers = self.eb_clers[self.eb_clers_offset]
        self.eb_clers_offset += 1

        if clers == 'C':
            O[N(c)] = -1
            self.eb_N += 1
            V[3*self.eb_T] = self.eb_N
        elif clers == 'L':
            O[N(c)] = -2
            if not _eb_check_handle(self, N(c)):
                _eb_zip(self, N(c))
        elif clers == 'R':
            O[c] = -2
            _eb_check_handle(self, c)
            c = N(c)
        elif clers == 'S':
            _eb_decompress_connectivity(self, c)
            c = N(c)
            if O[c] > 0:
                return
        elif clers == 'E':
            O[c] = O[N(c)] = -2
            _eb_check_handle(self, c)
            if not _eb_check_handle(self, N(c)):
                _eb_zip(self, N(c))
            return

def _eb_check_handle(self, c):
    def P(c): return c/3+(2+c)%3
    def N(c): return c/3+(1+c)%3
    N, O, H = self.eb_N, self.eb_O, self.eb_H
    A = self.eb_A

    if len(H) < 1:
        return 0

    if c != H[self.eb_A+1] or self.eb_A > len(self.eb_H):
        return 0
    else:
        O[c] = H[A]; O[H[A]] = c
        a = P(c)
        while O[a] >= 0 and a != H[A]:
            a = P(O[a])
        if O[a] == -2:
            _eb_zip(self, a)
        a = P(O[c])
        while O[a] >= 0 and a != c:
            a = P(O[a])
        if P[a] == -2:
            _eb_zip(self, a)
        self.eb_A += 2
        return 1

def _eb_zip(self, c):
    def P(c): return int(c/3)+(2+c)%3
    def N(c): return int(c/3)+(1+c)%3
    def T(c): return int(c/3)
    O, H, V = self.eb_O, self.eb_H, self.eb_V

    b = N(c)
    while O[b] >= 0 and O[b] != c:
        b = N(O[b])
    if O[b] != -1:
        return
    O[c] = b; O[b] = c
    a = N(c)
    V[N(a)] = V[N(b)]
    while O[a] >= 0 and a != b:
        a = N(O[a])
        V[N(a)] = V[N(b)]
    c = P(c)
    while O[c] >= 0 and c != b:
        c = P(O[c])
    if O[c] == -2:
        _eb_zip(self, c)

def _eb_decompress_vertices(self, c):
    def P(c): return int(c/3)+(2+c)%3
    def N(c): return int(c/3)+(1+c)%3
    def T(c): return int(c/3)
    O, H, V = self.eb_O, self.eb_H, self.eb_V
    U = self.eb_U; G = self.eb_G; M = self.eb_M

    while 1:
        U[T(c)] = 1
        if M[V[c]] != 1:
            self.eb_N += 1
            #G[sefl.eb_N] = _eb_decode_delta(self, c)
            G.append(_eb_decode_delta(self, c))
            M[V[c]] = 1
            c = O[N(c)] # c.r
        elif U[T( O[N(c)] )] == 1:
            if U[T( O[P(c)] )] == 1:
                return
            else:
                c = O[P(c)] # c.l
        elif U[T( O[P(c)] )] == 1:
            c = O[N(c)]
        else:
            _eb_decompress_vertices(self, O[N(c)])
            c = O[P(c)]
            if U[T(c)]:
                return

def _eb_decode_delta(self, c):
    return [float(x) for x in self.f_verts.readline().split()]
