#! /usr/bin/env python
#
#    Base model class
#    Copyright (C) 2002  Michael Urman
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

class Model:
    """Node base class for modeling system"""
    def __init__(self, *children):
        if children: self._children = children
        else: self._children = []

        self.draw_wires = 0
        self.draw_faces = 1
        self.draw_textures = 0
        self.draw_duals = 0
        self.draw_normals = 0

    def __repr__(self):
        return '<ModelNode id %s>'%id(self)

    def draw(self, *args, **kvargs):

        if self.draw_wires: kvargs['wires'] = 1
        if self.draw_faces: kvargs['faces'] = 1
        if self.draw_textures: kvargs['textures'] = 1
        if self.draw_duals: kvargs['duals'] = 1
        if self.draw_normals: kvargs['normals'] = 1

        for c in self._children: c.draw(*args, **kvargs)

    def add(self, *children):
        """add(self, children) -> None

        Add all children under the current Node"""

        self._children.extend(children)

    def save_ovtable(self, *args, **kvargs):
        """Pass on save_ovtable to interested children"""
        for c in self._children: c.save_ovtable(*args, **kvargs)

    def load_ovtable(self, *args, **kvargs):
        """Pass on load_ovtable to interested children"""
        for c in self._children: c.load_ovtable(*args, **kvargs)

    def save_edgebreaker(self, *args, **kvargs):
        """Pass on save_edgebreaker to interested children"""
        for c in self._children: c.save_edgebreaker(*args, **kvargs)

    def load_edgebreaker(self, *args, **kvargs):
        """Pass on load_edgebreaker to interested children"""
        for c in self._children: c.load_edgebreaker(*args, **kvargs)
